#include <cuda_fp16.h>

{{corr_enum}}

extern "C" {
	__global__ void correct(const unsigned short* data, // shape: memory cell, 1, ss, fs
	                        const unsigned short* cell_table,
	                        const unsigned char corr_flags,
	                        const unsigned short input_frames,
	                        const unsigned short map_memory_cells,
	                        const float* offset_map, // shape: cell, ss, fs, gain
	                        const float* gain_amp_map,
	                        const float* rel_gain_slopes_map,
	                        const float* flatfield_map,
	                        const unsigned int* bad_pixel_map,
	                        const float bad_pixel_mask_value,
	                        float* gain_map, // similar to preview for AGIPD
	                        {{output_data_dtype}}* output) {
		const size_t fs_dim = {{fs_dim}};
		const size_t ss_dim = {{ss_dim}};

		const size_t frame = blockIdx.x * blockDim.x + threadIdx.x;
		const size_t ss = blockIdx.y * blockDim.y + threadIdx.y;
		const size_t fs = blockIdx.z * blockDim.z + threadIdx.z;

		if (frame >= input_frames || ss >= ss_dim || fs >= fs_dim) {
			return;
		}

		const size_t data_stride_fs = 1;
		const size_t data_stride_ss = fs_dim * data_stride_fs;
		const size_t data_stride_frame = ss_dim * data_stride_ss;
		const size_t data_index = frame * data_stride_frame + ss * data_stride_ss + fs * data_stride_fs;
		const unsigned short raw_data_value = data[data_index];
		const unsigned char gain = (raw_data_value >> 12) & 0x0003;
		float corrected = (float)(raw_data_value & 0x0fff);
		float gain_for_preview = (float)gain;

		const size_t gm_map_stride_gain = 1;
		const size_t gm_map_stride_fs = 3 * gm_map_stride_gain;
		const size_t gm_map_stride_ss = fs_dim * gm_map_stride_fs;
		const size_t gm_map_stride_cell = ss_dim * gm_map_stride_ss;

		const size_t map_cell = cell_table[frame];
		if (map_cell < map_memory_cells) {
			const size_t gm_map_index = gain * gm_map_stride_gain +
				map_cell * gm_map_stride_cell +
				ss * gm_map_stride_ss +
				fs * gm_map_stride_fs;

			if (gain > 2 || ((corr_flags & BPMASK) && bad_pixel_map[gm_map_index])) {
				// now also checking for illegal gain value
				corrected = bad_pixel_mask_value;
				gain_for_preview = bad_pixel_mask_value;
			} else {
				// TODO: save GPU memory by combining maps
				if (corr_flags & OFFSET) {
					corrected -= offset_map[gm_map_index];
				}
				if (corr_flags & GAIN_AMP) {
					corrected *= gain_amp_map[gm_map_index];
				}
				if (corr_flags & REL_GAIN) {
					corrected *= rel_gain_slopes_map[gm_map_index];
				}
				if (corr_flags & FF_CORR) {
					corrected *= flatfield_map[gm_map_index];
				}
				if (corrected < -1e7 || corrected > 1e7 || isnan(corrected) || isinf(corrected)) {
					corrected = bad_pixel_mask_value;
				}
			}
		} else {
			if (gain > 2 || corrected < -1e7 || corrected > 1e7) {
				corrected = bad_pixel_mask_value;
				gain_for_preview = bad_pixel_mask_value;
			}
		}

		gain_map[data_index] = gain_for_preview;
		{% if output_data_dtype == "half" %}
		output[data_index] = __float2half(corrected);
		{% else %}
		output[data_index] = ({{output_data_dtype}})corrected;
		{% endif %}
	}
}
